// -*- C++ -*-
//
// Package:     Arich
// Module:	ArichParticle
//
// Description: 
//        This module desribes motion of particle in the
//        medium
//        It includes methods for 
// Usage:
//      [class and public member function
//       documentation goes here]
//
// Author:      R. Pestotnik
// Created:     Thu May  1 00:00:01 EST 2003
// $Id$
//
// Revision history
//
// $Log$


#include "ArichConfig.hh"
#include "ArichDetector.hh"
#include "ArichAerogel.hh"
#include "ArichGeo.hh"
#include "CLHEP/Random/Randomize.h"


ArichParticle::ArichParticle(){};
ArichParticle::~ArichParticle(){};

ArichParticle::ArichParticle(Hep3Vector r, Hep3Vector dir, HepDouble p,HepInt Lundtype){ Set(r,dir,p,Lundtype);}


void ArichParticle::Set(Hep3Vector r, Hep3Vector dir, HepDouble p,HepInt Lundtype){
  // Description: sets the initial parameters of the track

  const HepDouble p_mass[5]={.511E-3,105.66E-3,139.57E-3,493.68E-3,938.27E-3};
  fVertexPosition  = r;
  fVertexDirection = dir;
  fMomentum        = p;
  fMCtype          = Lundtype;
  HepInt type      = Lund2Type(fMCtype);
  if (fMCtype>0) fCharge=+1; else fCharge=-1;
  if (type<0) fMass = p_mass[0];
  else        fMass = p_mass[type];
  fNgenPhot = 0;
  fTracked  = 0;
  fRecoFlag = 0;
  fPath     = 0;

  for (int i=0;i<5;i++) {
    plkh[i]=0;
    stfs[i]=0;
  }
  
  for (int i=0;i<4;i++) fSigma[i]=0;
    
  stfb=0;
  sfot=0;
  
  return;
}


HepInt ArichParticle::Lund2Type(int ipart) {
  //Description: Converts GEANT code to e mu pi K p index.
  // return =0  electron
  //        =1  muon
  //        =2  pion
  //        =3  kaon
  //        =4  proton
  //        =-1   ?
  ipart=abs(ipart);
  switch (ipart) {
  case  11  :  return 0;
  case  13  :  return 1;
  case  211 :  return 2;
  case  321 :  return 3;
  case  2212:  return 4;
  default:     return -1;
  }
};

extern HepInt BzTrack(HepDouble mass,HepDouble charge,Hep3Vector r,
			  Hep3Vector dir, HepDouble momentum,
		      HepDouble Bz, HepDouble zout, HepDouble rcut,
			  Hep3Vector &r0, Hep3Vector &dir0  );

HepInt  ArichParticle::ChargedParticleToPlane(HepDouble Bz, HepDouble zout){
  // Description:
  // Propagation of the charged particle to the plane perpendicular to the
  // direction of the magnetic field which is in z direction
  //
  // Input Parameters:
  // Bz magnetic field density
  // zout z of the plane
  //
  // returns:
  // 1 if the particle wass succesfully propagated
  // 0 the particle escapes through the barrel 
 return  BzTrack(fMass,fCharge,fPosition,fDirection,fMomentum,Bz,zout,gArichDetector->fRadius[1],fPosition,fDirection);
 
}

HepInt    ArichParticle::Track(){
  // Description:
  // Propagation of the track to the photon detector
  // Output parameters:
  // fPosition, fDirection, fAeroIn,fAeroOut, fDetector are updated
  // Return:
  // fTracked=1 Tracking was succesful
  // fTracked=-1 Track is lost in the barrel
  // Calls: ChargedParticleToPlane, gArichAerogel->Inside
  // Called by: PID_Arich::ParticleTracking
  
  if ( fTracked ) return fTracked;


  fPosition  = fVertexPosition;
  fDirection = fVertexDirection;
  fTracked= -1;
  // check the direction
  if (fDirection.z()<0) return fTracked;

  HepDouble Bz= gArichDetector->fMagneticField;
  // intersection with aerogel entrance
  if (!ChargedParticleToPlane(Bz, gArichAerogel->zin())) return fTracked;
  if (!gArichAerogel->Inside(fPosition))                 return fTracked;
  fAeroIn = fPosition;
   // intersection with aerogel exit
  if (!ChargedParticleToPlane(Bz, gArichAerogel->zout())) return fTracked;
  fAeroOut = fPosition;
    
  // propagation to the detector  
  ChargedParticleToPlane(Bz, gArichDetector->z());
  fDetector = fPosition;

  return fTracked = 1;
}


HepInt    ArichParticle::CherenkovPhotonEmission(){
  // Description:
  // Radiation of Cherenkov Photons in the aerogel: approximation is used.
  // Photons are radiated linearly from entrance to exit of the particle in aerogel.
  //
  // Output Parameters:
  // fMeanEPoint, fMeanEDir :Mean emmision point and direction
  // fVirtualCenter :virtual intersection of particle direction with detector
  //
  // returns 
  // fNgenPhot number of photons generated.
  // Calls: TrackSystem,ArichPhoton::CherenkovPhotonEmmission
  fNgenPhot=0;
  if ( fTracked <= 0) return fNgenPhot;

  // good enough approximation
  fPath = (fAeroOut - fAeroIn).mag(); 
  //  mean emission point
  HepDouble trlen = gArichAerogel->TransmissionLength();
  HepDouble dla   = fPath/trlen;
  HepDouble dmean;
  if (dla > 0.01) {
    dmean=trlen*(1-dla/(exp(dla)-1));
  } else {
    dmean=fPath/2;
  }

  fMeanEPoint=fAeroOut-dmean*fDirection;
  fMeanEDir  =(fAeroOut - fAeroIn).unit();

  // virtual intersection of particle direction with detector
  fVirtualCenter = RayToPlane(fMeanEPoint,fMeanEDir,gArichDetector->fNorm,gArichDetector->fEntrance);
  
  TrackSystem();


  HepDouble beta     = Beta();
  HepDouble rind     = gArichAerogel->RefindMax();
  HepDouble sthc2    = 1- 1/sqr(rind*beta);
  HepDouble sthc2max = gArichAerogel->Sthc2Max();
  HepDouble Nmean    = gArichAerogel->N0()*fPath*sthc2;
  HepInt    Ngen     = ArichRandPoisson(Nmean);
  
  ArichPhoton fot;
  HepInt first=1;
  for (int i=0;i< Ngen;i++){
        
    Hep3Vector EmissionPoint  = fAeroIn+ ArichRand() *  fPath * fMeanEDir;
    HepInt     IsGenerated = fot.CherenkovPhotonEmission( EmissionPoint,  fMeanEDir, beta );
    
    if (IsGenerated) {
      gArichPhoton.push_back(fot);
      fNgenPhot++;
      if (first) {
	HepInt idx=gArichPhoton.size()-1; 
	fPhoton=idx;
	first=0;
      }
    }
    
  }
  return fNgenPhot;
}


HepInt ArichParticle::TrackSystem(){
  // Description: Calculates the base vectors of the track system
  // Output:
  // fX,fY,fZ base vectors
  
  HepDouble ss=1-sqr(fMeanEDir.y());
  if (ss>0){
    ss=sqrt(ss);
    fZ=fMeanEDir;
    fX=Hep3Vector(fZ.z()/ss,0,-fZ.x()/ss);
    fY=Hep3Vector(fZ.y()*fX.z(), fZ.z()*fX.x()-fZ.x()*fX.z(), -fZ.y()*fX.x());
  } else {
    fZ=Hep3Vector(0,1,0);
    fX=Hep3Vector(0,0,1);
    fZ=Hep3Vector(1,0,0);
  }
  return (ss>0);
}




HepDouble ArichParticle::SignalEstimate(){
  // Description:
  //  Estimatate the number of signal photons
  //
  //  Output parameters:
  // thc[o..nhyp]  expected cherenkov angle for the hypotheses
  // stfs[0..nhyp] number of signal photons
  // fRecoFlag: 0, if there is no expected photons for any mass hyp.,otherwise 100*acc(electron)
  // Calls:  ExpectedCherenkovAngle, GeometricalAcceptance
  // return
  // -1 track in wrong direction
  // 0  otherwise: 
  const HepDouble p_mass[5]={.511E-3,105.66E-3,139.57E-3,493.68E-3,938.27E-3};
  const HepInt    nmass  = 5;
  const HepDouble dthc   = 2.E-3;
  const HepDouble thc_min= 50.E-3;
  const HepDouble n_zero = 10;
  const HepDouble eta    = 0.37;
    

  fRecoFlag=0;
  if (fDirection.z()<0)  return -1;
  
  HepDouble tracklen   = gArichAerogel->Thickness()/fDirection.z();
  HepDouble refind =  gArichAerogel->Refind();
  HepDouble sig_pmt = gArichDetector->SigmaPad();
  // cerenkov angle of hypotheses
  HepInt hmin= min(NHYP,nmass);
  for (int i=0;i< hmin; i++ ){
    thc[i]= ExpectedCherenkovAngle(fMomentum,p_mass[i], refind);
  }

  // geometrical acceptance and number of photons
  HepDouble acc=0;
  HepDouble s=0;
  stfs[0]=0;
  for (int i=0;i< hmin; i++ ){
    acceptance[i]=0;
    if (i){
      if (fabs(thc[i]-thc[i-1])> dthc){
	acceptance[i]= GeometricalAcceptance(thc[i],thc_min,sig_pmt);
      } else {
	acceptance[i]= acceptance[i-1];
      }
    } else {
      acceptance[0]=GeometricalAcceptance(thc[0],thc_min,sig_pmt);
    }
    HepDouble s2thc=sin(thc[i]);
    s2thc*=s2thc;
    stfs[i]=n_zero*tracklen * eta * acceptance[i] * s2thc;
    //cout << this << " i:" << i << " " << stfs[i] << endl;
    s= max(s,stfs[i]);
  }
  if (s>0.0) fRecoFlag= (int) (100.*acceptance[0]);

  for (int i=nmass;i<NHYP;i++){
    thc[i]=thc[1];
    acceptance[i]=acceptance[1];
    stfs[i]=stfs[1]*(i-nmass+1);
  }

  return 0;
}  

HepDouble ArichParticle::BackgroundEstimate(){
  // Description:
  // Estimation of the number of background photons
  // Input parameters
  // fSab   narrow  windows linear integral
  //
  // returns:
  // stfb
  stfb = fSab*0.05*400;
  return stfb;
}

HepInt    ArichParticle::SetRecoWindow(){

  // Description:
  // Calculate the intervals for signal and background calculation
  // Output Parameters:
  // fNarrowWindow fWideWindow
  
  
  fSab=0;
  fAx = 0;         // normalisation parameter for background (linear)
  fAB = 0;         //  fit factor for background photons (linear)

  HepDouble a[NHYP],b[NHYP];
  HepInt nab;
  // narrow windows
  HepInt nhy = min (NHYP,6);
  HepDouble sig     = gArichDetector->SigmaPad();
  HepDouble sigma   = sqrt(sig*sig + (sqr(fSigma[0])+sqr(fSigma[1]))*.5);
 
  HepDouble thcmin=  gArichDetector->fThcMin;
  HepDouble thcmax=  gArichDetector->fThcMax;

  // Narrow windows
  HepDouble swin = gArichDetector->fNarrowWinSize * sigma;
  SetRWin(nhy,thc,swin,thcmin,thcmax,a,b,nab,fSab);

  for (int i=0;i<nab;i++) fNarrowWindow.push_back(ArichInterval(a[i],b[i]));
  
  // Wide windows
  HepDouble dwin = gArichDetector->fWideWinSize * sigma; 
  SetRWin(nhy,thc,dwin,thcmin,thcmax,a,b,nab,fSab);

  for (int i=0;i<nab;i++) fWideWindow.push_back(ArichInterval(a[i],b[i]));
  

  if (fSab > 0 ){
    fAx=1/fSab;
    fSf=sqr(thcmax)/2.*fAx;
    if (fScd-fSab) fAB=fSab/(fScd-fSab); else fAB=0;
  } else {
    fAx=0;
    fAB=0;
    fSf=0;
  }
  
  return 0;
}

HepInt    ArichParticle::SetRWin(HepInt nn, HepDouble  *thc, HepDouble win, HepDouble x1_min, HepDouble x2_max,HepDouble *x1, HepDouble *x2,HepInt &n, HepDouble &sum1){


  // Description:
  // Calculates the intervals for calculation of number of signal photons
  //
  // Input Parameters:
  // nn                   ! number of hypotheses
  // thc[0:NN-1]          ! cerenkov angle for other,e,mu,pi,K,p
  // win                  ! window width
  // x1_min               ! lower limit for X1
  // x2_max               ! upper limit for X2
  //
  // Output Parameters
  // sum1                 ! integral (lin.)
  // x1,x2                ! lower and upper limits of the interval
  // n                      number of interval
  // Called by : Likelihood
  n=0;  
  HepInt i=0;
  while ( thc[i]<=0 ) {
    i++ ;
    if (i==nn) return 0;
  }
  // cout << thc[i] << "SetRwin" << i << endl;
  x2[n]=min(thc[i]+win/2.,x2_max);
  x1[n]=min(max(thc[i]-win/2,0),x2_max);

  for ( int j=i+1;j<nn;j++ ){
    //cout << thc[j] << "SetRwin"<< n << " $" << j << endl;
    if (thc[j] > 0){
      HepDouble x=thc[j]+win/2.;
      if ( x <x1[n] ) x2[++n]=x;
      x1[n]=max(thc[j] - win/2., x1_min);
    }
  }
  n++;
  //  sum=0;
  sum1=0;
  for ( int j=0;j<n;j++){
    sum1+= sqr(x2[j]) - sqr(x1[j]);
  }
  sum1/=2;
  return 0;
}

HepInt ArichParticle::IsInNarrowWindow(HepDouble thc){
  HepInt dum=0;
  for (int i=0;i<fNarrowWindow.size();i++){
    if ((thc<fNarrowWindow[i].ul) && (thc>fNarrowWindow[i].ll)) return 1;
  }
  return 0;
}


HepInt ArichParticle::IsInWideWindow(HepDouble thc){
  HepInt dum=0;
  
  for (int i=0;i<fWideWindow.size();i++){
    if ((thc<fWideWindow[i].ul) && (thc>fWideWindow[i].ll)) return 1;
  }
  return 0;
}

HepDouble ArichParticle::Beta(){
  // Description:
  // Calculates velocity of the particle from its momentum and mass
  HepDouble pm=fMomentum/fMass;
  pm*=pm;
  return sqrt(pm/(1+pm));
}

HepDouble  ArichParticle::ExpectedCherenkovAngle(HepDouble p,HepDouble mass, HepDouble refind){
  // Description:
  // calculates Cerenkov angle of a particle of mass MASS and momentum P
  //
  // Input Parameters:
  // p  momentum of the particle
  // mass of the particle
  //
  // return expected cherenkov angle
  
  if (mass<=0 || p<0 ) return 0;
  HepDouble a = p/mass;
  HepDouble b = a/sqrt(1+a*a)*refind;
  if (b>1) return acos(1/b);
  return 0;
};


HepInt ArichParticle::CerenkovAngle(Hep3Vector  hit,HepDouble &th_cer,HepDouble &fi_cer){
  // Description:
  // Reconstruction of  the cherenkov polar and azimuthal angle from known hit
  //
  // Input Parameters:
  // hit position of the hit on the detector
  //
  // Output Parameters:
  // th_cer, fi_cer polar and azimuthal angle
  // Called by ArichLikelihood
  // Calls CerenkovPhoton

  // Arguments:
  // Input: hit
  // Output: the cherenkov polar and azimuthal angle
  
  Hep3Vector rf=fAeroOut;
  Hep3Vector dirf;

  HepInt niter= CerenkovPhoton(fMeanEPoint,fMeanEDir,hit,rf,dirf);

  Hep3Vector dirch=Hep3Vector(dirf*fX,dirf*fY,dirf*fZ);
  //cout << acos(((hit-fMeanEPoint).unit()*fMeanEDir))<< "acos"  << rf << endl;
  //cout << "CerenkovAngle" <<fMeanEPoint << fMeanEDir << hit << rf << dirf << dirch;
  th_cer=acos(dirch.z());
  fi_cer=dirch.getPhi();
  //cout << " thc=" << th_cer << " niter" << niter<<  endl;
  return niter;
}

inline Hep3Vector ArichParticle::Refraction(Hep3Vector s, HepDouble n){
  HepDouble x=s.x()/n;
  HepDouble y=s.y()/n;
  //cout << s << x  << "::" << y << endl;
  return Hep3Vector(x,y,sqrt(1-x*x-y*y));
}


HepInt ArichParticle::CerenkovPhoton(Hep3Vector r, Hep3Vector dir, Hep3Vector rh,
				Hep3Vector &rf, Hep3Vector &dirf){
  //
  // Description:
  // The method calculates the direction of the cherenkov photon
  // and intersection with the aerogel exit point
  //
  // Called by: CerenkovAngle

  // Arguments:
  // Input:
  // dir, r track position and direction
  // rh photon hit
  
  // Output:
  // rf aerogel exit
  // dirf photon direction in aerogel
  
  //static HepDouble dwin = gArichDetector->fWindowThickness;

  const HepDouble dmin = 0.01;
  const HepInt    niter= 100;
  //Hep3Vector dirw;
  //Hep3Vector rh1=rh;
  //rh1(2)=rh1(2)-dwin;
  HepDouble dz = gArichAerogel->z()-r.z();
  HepDouble s;
  for (int iter=0;iter<niter;iter++){
    Hep3Vector dirl=(rh-rf).unit();
    
    dirf=Refraction(dirl, gArichAerogel->Refind() );
    //dirw=Refraction(dirl, gArichDetector->fWindowRefind );
    HepDouble x1=rf.x();
    HepDouble y1=rf.y();
    rf.setX(r.x()+ dirf.x()/dirf.z()*dz);
    rf.setY(r.y()+ dirf.y()/dirf.z()*dz);
    HepDouble d2= sqr(rf.x()-x1) + sqr(rf.y()-y1);
    //cout << iter  << dirl << dz <<  "CerenkovPhoton"<< dirf << rf << " d2 " << d2<< endl;
    if ( ( d2 < dmin ) && (iter)) return iter;
  }
  return -1;
}



HepDouble  ArichParticle::GeometricalAcceptance(HepDouble thc,HepDouble thc_min,HepDouble sigma){

  //
  // Description: The method calculates the Geometrical Acceptance of the track
  //             
  // Called by: ArichLikelihood

  // Arguments:
  // Input:
  // thc cerenkov angle
  // thc_min
  // sigma 
  
  // Output:
  // rf aerogel exit
  // dirf photon direction in aerogel

  return 1;
  if (thc < thc_min) return 0;
  
  Hep3Vector r    = fMeanEPoint;
  HepDouble  sthc = sigma/1.41;

  const HepInt ngen=100;
  

  Hep3Vector dirf, dirf0,rf;
  HepInt iacc=0,xch,ych,ier;
  
  for (int i=0;i < ngen; i++){

    dirf0=Direction( RandGauss::shoot(gRanluxEngine,thc,sthc),
		     2*M_PI*ArichRand()); // photon in track system

    dirf =Rotation(fDirection,dirf0);                 // photon in local
    rf=RayTrace(r,dirf);
    //cout << " geoacc " << rf << endl;

    ier=gArichDetector->Position2Hit(rf,xch,ych);
    if (ier<0) continue;
    iacc++;
  }
  return (float) iacc/ngen;
 
};


ostream&operator<<(ostream& o, ArichParticle& d){
  if (d.fTracked>0)   return o << " Track PID=" << d.fMCtype << ">" << d.Lund2Type(d.fMCtype)
			       << " npgen=" << d.fNgenPhot  
			       << " r="   << d.fPosition << " mm" 
			       << " dir=" << d.fDirection
			       << " p="   << d.fMomentum << " GeV/c"
			       << " ain="   << d.fAeroIn
			       << " emp="   << d.fMeanEPoint
			       << " aout="  << d.fAeroOut
			       << " virtual="   << d.fVirtualCenter
			       << " tracked="  << d.fTracked
			       << " RECO="     << d.fRecoFlag
			       << " ID="       << d.fPid;
  else  return o << " Track PID=" << d.fMCtype << ">" << d.Lund2Type(d.fMCtype)
		 << " tracked="  << d.fTracked;
    
}













