// -*- C++ -*-
//
// Package:     Arich
// Module:	ArichPhoton
//
// Description: The class implements methods for photon emmision and tracking
//
// Usage:
//      [class and public member function
//       documentation goes here]
//
// Author:      R. Pestotnik
// Created:     Thu May  1 00:00:01 EST 2003
// $Id$
//
// Revision history
//
// $Log$


#include "ArichConfig.hh"
#include "ArichPhoton.hh"
#include "ArichAerogel.hh"
#include "ArichDetector.hh"

ArichPhoton::ArichPhoton(){}
ArichPhoton::~ArichPhoton(){}


HepInt  ArichPhoton::CherenkovPhotonEmission(Hep3Vector r,Hep3Vector dir, HepDouble beta){

  // Description: generation of the Cherenkov Photon
  //
  // Input parameters: parameters of the  track
  // Hep3Vectpr r,dir particle position and momentum at the emmission point
  // HepDouble  beta     particle velocity
  //
  // Returns:
  // 0 photon is not emmited
  // 1 photon is emmited
  //
  // Calls: CalculatePhoton
  //
  static const HepDouble ef[2]={gArichDetector->Energy(0),gArichDetector->Energy(1)};
  static const HepDouble fSthc2Max= gArichAerogel->Sthc2Max();
   
  //cout << "ArichPhoton::CherenkovPhotonEmission" << endl;
  
  fEnergy = ef[0]+(ef[1]-ef[0])*ArichRand();
  if(!gArichDetector->DetectorQE(fEnergy)) return 0;
  
  // dispersion
  HepDouble fRind    = gArichAerogel->Refind(fEnergy);
  HepDouble fRindMax = gArichAerogel->RefindMax();
  HepDouble cthc2max = 1/sqr(beta*fRindMax);
  HepDouble cthc2    = 1/sqr(beta*fRind);
  HepDouble probability=(1-cthc2)/(1-cthc2max);
  if (ArichRand() > probability) return 0;

  fFic= 2*M_PI*ArichRand();
  fThc= acos(sqrt(cthc2));

  fPosition=r;
  
 
  CalculatePhoton(fThc,fFic,r,dir,fDirection,fPolarization);
 
  return 1; 
}


void ArichPhoton::CalculatePhoton(HepDouble thc,HepDouble fic,Hep3Vector r,Hep3Vector dir,
				   Hep3Vector &dirf,Hep3Vector &polf){

  // Description:
  // From known azimuthal and polar angle calculates the direction and
  // the polarization of the Cherenkov photon
  //
  // Input parameters:
  // thc,fic   polar and azimuthal angle of the photon
  // r,dir particle position and direction
  //
  // Output parameters:
  // dirf photon direction
  // polf photon polarization
  //
  // Called by: CherenkovPhoton
  
  dirf=Direction(thc,fic);  // particle system
  dirf=Rotation(dir,dirf);  // global system

  // photon polarization
  HepDouble thpol=M_PI/2+thc;
  polf=Direction(thpol,fic);  // particle system
  polf=Rotation(dir,polf);    // global system

}



HepDouble ArichPhoton::Track(){
  //Description:
  //Tracking of the photon from the emmission poit to the photon detector
  //
  // Input/Output Parameters:
  // fPosition, fDirection, fPolarization, fPath  are updated
  // Called by: PID_Arich::PhotonTracking
  // Calls: RayleighScattering, ReflectionRefraction
  
  // check if is inside aerogel
  // cout << "ArichPhoton::Track()" << endl;

  fPath=0;
  if (!gArichAerogel->Inside(fPosition)) return fTracked=-2;

  HepDouble Rscatlen = gArichAerogel-> RayleighScatteringLength(fEnergy);
  HepDouble Rind     = 1/gArichAerogel->Refind(fEnergy);

  // calculate new scattering length
  HepDouble PathRayl = -Rscatlen * log (1- ArichRand());

  Hep3Vector norm;
  HepDouble d;
  HepInt type,nproc=0,iproc;

  HepDouble path;
  HepDouble path_all=0;
  //  PathRayl=100000;
  do {

    //cout << "ArichPhoton::Track" << endl;
    path=gArichAerogel->NearestBorder(fPosition,fDirection,norm, type);
    // cout << "nearest nproc=" << nproc << " path=" << path << "type=" << type;
    if (path> PathRayl){
      // Do Rayleigh Scattering
      // cout << "PathRayl" << PathRayl<< endl;
      fPosition=fPosition+fDirection*PathRayl;
      path_all+= PathRayl;
      RayleighScattering(fDirection,fPolarization);
      PathRayl = -Rscatlen * log (1- ArichRand());
      iproc=3;
    } else {
      fPosition+= fDirection*path;
      path_all += path;
      PathRayl -= path;

      iproc = ReflectionRefraction(fDirection,fPolarization,norm,Rind);

      // cout << "RR" << iproc<< fDirection << fPosition << endl;
	    
    }
    // increase the counter for number of processes occured
    nproc++;

  } while (iproc >0  && nproc < 5);
  
  // iproc == 3 Rayleigh
  // iproc == 2 total reflection
  // iproc == 1 reflection
  // iproc == 0 refraction

  // check if the photon is refracted in direction of the photon detector,
  // otherwise return 
  if (type || iproc) {
    fPath=-fPath;
    return fTracked=0;
  }

  // propagate photon to the detector
  path=RayPlaneDistance(fPosition,fDirection,gArichDetector->fNorm, gArichDetector->fEntrance);
  fPosition=fPosition+fDirection*path;
  fPath+= path;
  fTracked=1;
  return fPath;
  
}

ostream&operator<<(ostream& o, ArichPhoton& d){
  return o << " Photon E=" << d.fEnergy <<" eV "   
	   << " r="   << d.fPosition << " mm" 
	   << " dir=" << d.fDirection<< d.fDirection.mag() ;

}























