// -*- C++ -*-
//
// Package:     Arich
// Module:	ARichMain
//
// Description: 
//      The module implements Particle identuification in the aerogel RICH
// Usage:
//     
//     
//
// Author:      R. Pestotnik
// Created:     Thu May  1 00:00:01 EST 2003
// $Id$
//
// Revision history
//
// $Log$


#include "ArichConfig.hh"
#include "PID_Arich.hh"

#include "ArichDetector.hh"
#include "ArichAerogel.hh"
#include "ArichTube.hh"

ArichChrono   *gTimer[NGTIMER];
ArichAerogel  *gArichAerogel;
ArichDetector *gArichDetector;
ArichTube     *gArichTube;
RanluxEngine  *gRanluxEngine;
vector <ArichPhoton>   gArichPhoton;
vector <ArichParticle> gArichParticle;
vector <ArichHit>      gArichHit;
vector <ArichHitTab>   gArichHitTab;


PID_Arich::PID_Arich(){
  init(0,0);
}

PID_Arich::~PID_Arich(){
  //  for (int i=0;i<6;i++) if (fTimer[i]) delete fTimer[i];
  DEBUG(2) cout << "~PID_Arich()" << endl;
   Clear();
}


HepInt ArichDebugLevel=0;

HepInt PID_Arich::init(HepInt debug,HepInt mode){
  
  DEBUG(2) cout << "PID_Arich::Init" << endl;

  // debug level
  // 0 - nothing
  // 1 - histogramming
  // 2 - event output
  // 3 - track output
  // 4 - photon output

  // mode
  // 0 Full PID - (particle tracking, hit generation, ch.angle reconstruction, likelihood calculation)
  // 1 Fast PID
  
  
  ArichDebugLevel=debug;
  
  gRanluxEngine = new  RanluxEngine(19780503,4);
  
  gArichDetector  = new ArichDetector();
  gArichAerogel   = new ArichAerogel(); // aerogel should be initialized after detector

  gArichTube      = new ArichTube(gArichDetector->fPosition, gArichDetector->fNorm, gArichDetector->fRadius);

  DEBUG(2) {
    cout << *gArichAerogel  << endl;
    cout << *gArichDetector << endl;
    cout << *gArichTube << endl;
  }
  
  // allocate timers
  

  gTimer[NGTIMER-1]=new ArichChrono("Arich");

  // alocate memory
  
  gArichHit.reserve(2000);
  gArichParticle.reserve(50);
  gArichHitTab.reserve(10000);
  gArichPhoton.reserve(10000);


  fRecoTracks=0;
  fTrackedTracks=0;
  fRichTracks=0;
  fAllTracks  =0;
  fRichEvents=0;
  fAllEvents  =0;

  fFastPID = mode;
  return 0;  
}


HepInt PID_Arich::Run(){
  fAllEvents++;

  Clear();
  gTimer[NGTIMER-1]->start();

  switch (fFastPID) {
  case 0:
  case 1:
    DT(1,  ParticleTracking()   );
    DT(2,  SignalEstimate()     );
    DT(3,  BackgroundEstimate() );
    
    if (fFastPID) {
      DT(9, FastPhotonGeneration() );
    } else {
      DT(4, PhotonGeneration() );
      DT(5, PhotonTracking()   );  
      DT(6, Digitization()     );
      //HitReconstruction();
      DT(7,  CherenkovAngleReconstruction());
    }

    DT(8, Likelihood());

        
    fRichEvents+=( gArichHitTab.size()>0);
    break;
  case 2:
    DT(1,FastParticleId());
    break;

  default:
    cout << "This mode of Arich Simulation is not yet implemented\n";
    
  }
  
  gTimer[NGTIMER-1]->stop();
  DEBUG(2)  cout << fAllEvents
		 << " Buffers Photon " << gArichPhoton.size()
		 << " Hit " << gArichHit.size()
		 << " HitTab " << gArichHitTab.size()<<endl;
  return 0;
}


HepInt PID_Arich::End(){
  cout<<"Timer   "
      <<"N_measurements"<<" \t"
      <<"total"<<"    "
      <<"user"<<"    "
      <<"system"<<"    "
      <<"<total>"<<"    "
      <<"<user>"<<"    "
      <<"<system>"<<endl;
  
  for (int i=0;i<NGTIMER;i++) {
    if (gTimer[i]) gTimer[i]->printLine();
    //cout << endl;
  }
  if (fAllTracks) cout << (float) (fRichTracks)/ fAllTracks *100 << "%";
  cout << "\t Tracks with photons " << fRichTracks
       << "  all tracks " << fAllTracks << endl;

  if (fAllTracks) cout << (float) (fTrackedTracks)/ fAllTracks*100 << "%";
  cout << "\t Tracks in RICH " << fTrackedTracks
       << "  all tracks " << fAllTracks << endl;

  if (fAllTracks) cout << (float) (fRecoTracks)/ fAllTracks*100 << "%";
  cout << "\t Tracks Identified " << fRecoTracks
       << "  all tracks " << fAllTracks << endl;
  
  if (fAllEvents)  cout << (float) (fRichEvents)/ fAllEvents*100<< "%";
  cout << "\t Events with photons " << fRichEvents
       << "  all events " << fAllEvents << endl;
  
  DEBUG(2) cout << "PID_Arich::End" << endl;
  
  return 0;
}


HepInt PID_Arich::FastPhotonGeneration(){
  //   aerogel
  const HepDouble sigma0= 0.013;
  const HepDouble bin_size=0.0015;
  const HepInt    nbin=256;
  
  for (int i=0;i<gArichParticle.size();i++){
  

    if (gArichParticle[i].fRecoFlag<=0) continue;
    // background    
    HepInt bgrfot = ArichRandPoisson(gArichParticle[i].fSf*gArichParticle[i].stfb);      // generated number of photons

    // calculate number of detected photons (input ndet=n0*l*sin2thc -> genfot)
    HepInt id     = gArichParticle[i].MCIdentity();
    HepInt genfot = ArichRandPoisson(gArichParticle[i].stfs[id]);
    HepDouble thc = gArichParticle[i].thc[id];    

    gArichParticle[i].fNgenPhot = genfot;
    
    float wt=1, th_cer;
    float thcmax = 0.4;
    DEBUG(2) cout << "Gen: thc "<< thc  << " GenFot=" << gArichParticle[i].stfs[id] << " P:"  << genfot << " BgrFot="<<  bgrfot << " " << gArichParticle[i].stfb << " ID=" << id << endl;
    int kmax=genfot+bgrfot;
    HepDouble sigma;
    
    for (int k=0; k< kmax;k++){
      // generate ch.angle for genfot+bgr  photons
      if ( k<genfot ) th_cer= Arich::RandGauss(thc,sigma0);
      else            th_cer= Arich::RandBackground(thcmax);
      //cout <<   gArichParticle[i] << endl;
      if (gArichParticle[i].IsInWideWindow(th_cer)){

	DEBUG(2) cout << k << " :" << th_cer << " GenFot=" <<  genfot << " BgrFot="<<  bgrfot << endl;      
        if (gArichParticle[i].IsInNarrowWindow(th_cer)) sigma=sigma0; else  sigma=-sigma0;
	HepInt ir= (int) (th_cer/bin_size);
	if ( ir < nbin  ) gArichHitTab.push_back(ArichHitTab(i,-1,sigma0,ir,th_cer,0,wt));
	} 
    }

  }
  return 0;
}
  
HepInt PID_Arich::PhotonTracking(){
  HepDouble path;
  // return 0;
  for (int i=0;i<gArichPhoton.size();i++){
    path = gArichPhoton[i].Track();
    DEBUG(4) cout << i << "path= " << path << gArichPhoton[i] << endl;
  }
  return 0;
}

HepInt PID_Arich::Digitization(){
  HepInt nhit=0,nh;
  for (int i=0;i<gArichPhoton.size();i++){
    if (gArichPhoton[i].fPath<=0) continue;
    nh = gArichDetector->Digitization(&gArichPhoton[i],1);
    if (nh>=0)  nhit ++; 
  }
  DEBUG(2) cout << "Digitization() end -> hits=" << nhit << " photons=" << gArichPhoton.size() << endl;

  return 0;
}

HepInt PID_Arich::ParticleTracking(){
  
  for (int i=0;i<gArichParticle.size();i++){
    if( gArichParticle[i].Track()>0 ) fTrackedTracks++;    
  }
  fAllTracks+=gArichParticle.size();
  return 0;
}


HepInt PID_Arich::PhotonGeneration(){

  for (int i=0;i<gArichParticle.size();i++){
    gArichParticle[i].CherenkovPhotonEmission();
    DEBUG(3) cout << i << gArichParticle[i] << endl;
  }
  return 0;
}


HepInt PID_Arich::GetEvent(){

  char *path="/u/argus/bracko/mcdata";
  char *filename="mc_gen_highmom";
  // char *filename="mc_gen";
  char *fileext="dat";
  static HepInt fileno=0;

  static char  fname[200];

  
  gArichParticle.clear();
  
  if (!fin) {
    sprintf(fname,"%s/%s_%d.%s",path,filename,fileno,fileext);
    fin=fopen(fname,"r");
    if (!fin) return -1;
    DEBUG(1) cout << fname << " opened"<< endl;
  }
  int j=0;
  HepFloat r[3],p[3];
  HepInt code,neve,ntra;
  int ndim=400;
  char line[ndim];
  int nb1=0;
  do {
    if (fgets(line,ndim,fin)==NULL) break;
    nb1=sscanf(line,"* %d%d",&neve,&ntra);
    //cout << nb1 << " " << line << endl;
    if (nb1==2){
      
      for (int i=0;i<ntra;i++){
	if (fgets(line,ndim,fin)==NULL) break;
	int nb2=sscanf(line,"%d%f%f%f%f%f%f",&code,&p[0],&p[1],&p[2],&r[0],&r[1],&r[2]);

	if (nb2==7) {
	  Hep3Vector pos(r[0],r[1],r[2]);
	  Hep3Vector mom(p[0],p[1],p[2]);
	  Hep3Vector dir    =mom.unit();
	  HepDouble momentum=mom.mag();
	  //momentum=2;
	  //code=+11;
	  gArichParticle.push_back( ArichParticle(pos,dir,momentum, code ) );
	}
      }
      DEBUG(2) cout << endl << "GetEvent " << neve << " ntracks=" << gArichParticle.size()<<endl;
    }
    if (feof(fin)) break; 
	
  } while (nb1!=2);
  if (feof(fin)) {
    fclose(fin);
    fin=NULL;
    DEBUG(1) cout << fname << " closed"<< endl;
    fileno++;
  }

  return neve;
}


HepInt   PID_Arich::pid( HepLorentzVector p, Hep3Vector p_measured, Hep3Vector x ){
  const float p_mass[5]={.511E-3,105.66E-3,139.57E-3,493.68E-3,938.27E-3};
 gArichParticle.clear();

  HepDouble   momentum=  p_measured.mag();
  Hep3Vector dir      =  p_measured.unit();
  HepDouble mass      =  p.m();
  HepInt type=0;
  for (int i=0;i<5;i++){
    if (mass <= p_mass[i]) { type=i; break;}
  }
  ArichParticle track;

  track.Set(x,dir,momentum, type );
  gArichParticle.push_back(track);
  return 0;
}


HepDouble  PID_Arich::prob(int type){
  if (gArichParticle.size()>0) return gArichParticle[0].plkh[type];
  else return -1;
}

HepInt PID_Arich::GetTracks( ){

  nTracks=1;
  for (int i=0;i<nTracks;i++){
    Hep3Vector r=gArichAerogel->fPosition0;
    Hep3Vector dir=Hep3Vector(0,0,1);
    HepDouble  p=3;
    HepInt type=2;
    ArichParticle track;
    track.Set(r,dir,p, type );
    gArichParticle.push_back(track);
    
    DEBUG(3) cout << "Track " << i<< endl;
  }
  return 0;
}



HepInt PID_Arich::HitReconstruction( ){
  gArichDetector->HitReconstruction();
  return 0;
}

HepInt PID_Arich::FastParticleId(){

  double plkh[5],stfs[5],thce[5],stfb,sfot,gfot;
  for (int i=0; i< gArichParticle.size(); i++){
    int id=gArichParticle[i].MCIdentity();
    float p= gArichParticle[i].fMomentum;
    //gArichParticle[i].fTracked=1;
    
    
    for (int j=0;j<5;j++) plkh[j]=0;


    FastPid(p,id,plkh,gfot,sfot,stfb,stfs,thce);

    
    gArichParticle[i].sfot     =sfot;
    gArichParticle[i].stfb     =stfb;
    gArichParticle[i].fNgenPhot= (int) gfot;
    gArichParticle[i].fRecoFlag=1;
    gArichParticle[i].fTracked =1;
    
    for (int j=0;j<5;j++) {
      gArichParticle[i].plkh[j]=plkh[j];
      gArichParticle[i].stfs[j]=stfs[j];
      gArichParticle[i].thc[j]=thce[j];
      gArichParticle[i].acceptance[j]=0;
    }
  }
  return 0;
}

HepFloat   PID_Arich::FastBackground(){
  // The parametrization of the background
  // based on the position of the track
  
  return 0.2;
}

HepInt PID_Arich::FastPid(float p, int id, double *plkh, double &genfot, double &sfot, double &stfb, double *stfs, double *thce){
  //   aerogel
  const float rind  = 1.05;
  const float n0    = 5;
  const float sigma0= 0.013;
  float l=20;//mm

  // particle
  const int nmass=5;
  const float p_mass[5]={.511E-3,105.66E-3,139.57E-3,493.68E-3,938.27E-3};
  double ax1=0.001;

  for (int i=0;i<nmass;i++){
    
    float pm=p/p_mass[i];
    //pm*=pm;
    thce[i] = sqrt(1+pm*pm)/(pm*rind);             // expected cosine of the cherenkov angle
    if (thce[i] > 1) thce[i]=1;
    stfs[i] = n0* l* (1-thce[i]*thce[i]);         // expected number of photons
    thce[i] = acos(thce[i]);                      // expected Cherenkov angle
    //cout << i<< " " << thce[i] << endl;
  }
  
  // backround
  float   thcmax = 0.5;
  stfb = FastBackground(); // mean number of background photons for (thc<thcmax)
  int nbgr = ArichRandPoisson(stfs[id]);      // generated number of photons

  // calculate number of detected photons (input ndet=n0*l*sin2thc -> genfot)
  genfot = ArichRandPoisson(stfs[id]);
  sfot=0;
  //if (genfot+nbgr) cout << id << " p=" << p << "genfot" << genfot << " stfb " << stfb << " nbgr " <<nbgr << "  ndet " << stfs[id] << " sfot " << sfot << endl;
  float thc=0; // particle cherenkov angle
  float wt=1;
  register float th_cer;
  
  for (int k=0; k< genfot;k++){
    // generate ch.angle for genfot photons
    th_cer= Arich::RandGauss(thce[id],sigma0);
    thc += th_cer;
    //    Hfill(170+id,p,th_cer,1 );   
    Arich::LogLikelihood(th_cer,sigma0,wt,nmass,thce,stfs,stfb,ax1, plkh, sfot);
    //cout << sfot << " S" <<th_cer << "mean=" << thce[id] << "sigma="<< sigma0 << endl;
 }

  for (int k=0; k< nbgr;k++){
    th_cer= Arich::RandBackground(thcmax);
    thc += th_cer;
    //Hfill(170+id,p,th_cer,1 );   
    Arich::LogLikelihood(th_cer,sigma0,wt,nmass,thce,stfs,stfb,ax1, plkh, sfot);
    //cout << sfot << " B" <<th_cer <<  endl;
  }

  // add poissonian part
  for (int k=0; k< nmass;k++) {
    plkh[k]+= Arich::PoisLog(stfs[k]+stfb,sfot);
    //if (genfot+nbgr) cout << plkh[k]<< " "  <<stfs[k] << " " << stfb << " "<< sfot << " " << endl;
  }

   

  return 0;

}



HepInt PID_Arich::CherenkovAngleReconstruction(){
  DEBUG(2) cout << "nHits" << gArichHit.size()<<" nTracks" << gArichParticle.size() << endl;
  
  const HepDouble bin_size=0.0015;
  const HepInt    nbin=256;
	


  ArichParticle *t;
  ArichHit      *h;
  HepDouble th_cer,fi_cer,sig,ut;
  HepInt ier;
  //  if ( gArichHit.size()     ==0) return 0;
  if ( gArichParticle.size()==0) return 0;


  for (int i=0; i< gArichParticle.size(); i++){
    t =  &gArichParticle[i];

    if (t->fRecoFlag == 0.0) continue;
    
    for (int ifot=0; ifot< gArichHit.size(); ifot++){
      h=&gArichHit[ifot];
       ier = t->CerenkovAngle(h->fPosition,th_cer,fi_cer);
       //cout << ier << " ArichPid::theta=" << th_cer << " fi=" << fi_cer << endl;
      if (ier < 0) continue;
      sig=0.013;
 
      
      if (t->IsInWideWindow(th_cer)){
	HepInt ir= (int) (th_cer/bin_size);
	if ( h->utez > 0 && ir < nbin  ) gArichHitTab.push_back(ArichHitTab(i,ifot,sig,ir,th_cer,fi_cer,h->utez));
      }       

    }

  }

  // BackgroundFitEstimate(ht,track, ntra);
  return 0;
}


HepInt PID_Arich::SignalEstimate(){
  for (int i=0; i< gArichParticle.size(); i++){
    if (gArichParticle[i].fTracked>0) {
      gArichParticle[i].SignalEstimate();
      if (gArichParticle[i].fRecoFlag > 0.0) gArichParticle[i].SetRecoWindow();
    }
  }
  return 0; 
}
 

HepInt PID_Arich::BackgroundEstimate(){
  for (int i=0; i< gArichParticle.size(); i++){
    if (gArichParticle[i].fRecoFlag>0)  gArichParticle[i].BackgroundEstimate();
  }
  return 0; 

}


HepInt PID_Arich::BackgroundFitEstimate(){

  ArichParticle *t=&gArichParticle[0];
  // reset;  
  for (int i=0; i< gArichParticle.size(); i++){
    t->stf_anti=0;
    t++;
  }

  //  count (photon with sig<0 is in antiwindow)
  ArichHit *h;
  HepDouble sig,th_cer,fi_cer,wt; 
  for (int i=0; i< gArichHitTab.size(); i++){
    t  = &gArichParticle[gArichHitTab[i].fTrack];
    wt = gArichHitTab[i].fWeight;
    if (sig < 0) t->stf_anti += wt;
  }

  //  estimate (do fit)
  t=&gArichParticle[0];
  for (int i=0; i< gArichParticle.size(); i++){
    t->stfb = t->stf_anti * t->fAB ;
    t++;
  }
  return 0;
};

HepInt PID_Arich::Clear( ){
   
  gArichPhoton.clear();
  gArichHit.clear();
  gArichHitTab.clear();
  
  gArichDetector->Clear();
  return 0;
}



HepInt PID_Arich::Likelihood(){
  
  ArichParticle  *t=&gArichParticle[0];
  
  

  // reset fields first
  for (int i=0; i< gArichParticle.size();i++){
    for (int k=0; k< NHYP;k++) t->plkh[k]=0;
    t->sfot=0;
    t++;
  }

  ArichHitTab *h=&gArichHitTab[0];
 
  ArichParticle  *tr;
  HepDouble sig,th_cer,fi_cer,wt;
  for (int k=0; k< gArichHitTab.size();k++){
    h->GetEntries(sig,th_cer,fi_cer,wt);
    tr =  &gArichParticle[h->fTrack];

    if (sig>0) Arich::LogLikelihood(th_cer,sig,wt,NHYP,tr->thc,tr->stfs,tr->stfb,tr->fAx, tr->plkh, tr->sfot);
    //    cout << k << " " << h->fTrack << " " << sig  << " Log " ;
    //for (int l=0; l< NHYP;l++) cout << " " << tr->plkh[l];
    //cout << endl;
    
    h++;
  }

  // add poissonian part, normalize it
  HepDouble stfe;
  for (int i=0; i< gArichParticle.size();i++){
    t=&gArichParticle[i];
    if (t->fNgenPhot > 0) fRichTracks++;
    if (t->fRecoFlag != 0.0 ) {
      fRecoTracks++;

      //cout << i << " Pois ";
      for (int k=0; k< NHYP;k++) {
	stfe=t->stfs[k]+t->stfb;
	t->plkh[k]+= Arich::PoisLog(stfe,t->sfot);
	//cout  << "#" << t->stfs[k] << " " <<t->stfb << " " << t->sfot << " " << t->plkh[k]; 
      }
      //cout << endl;
      //Normalize(t->plkh,NHYP,NHYP,t->fPid);
    }
    
  };
  return 0;
}




void PID_Arich::Normalize(HepDouble *plkh, HepInt nhyp, HepInt norm, HepInt &idp){
  //  plkh              ! log Likelihoods (on input and output) [in,out]
  //  nhyp              ! number of all hypotheses [in]
  //  norm              ! number of hypotheses to normalise them to sum=1 [in]
  //  idp               ! most probable particle ID [out]
  
  int n=min(norm,nhyp);
  idp=0;
  HepDouble pma=plkh[idp];
  if (n<=0) return;
  for (int i=0;i<n;i++){
    if(plkh[i]>pma) {
      pma = plkh[i];
      idp=i;
    }
  }  
  for (int i=0;i<nhyp;i++)  plkh[i]=plkh[i]-pma;
  HepDouble c=0;
  for (int i=0;i<n;i++) c+=exp(plkh[i]);
  c=log(c);
  for (int i=0;i<nhyp;i++)  plkh[i]=plkh[i]-c; // EXP(PLKH()) is probability
  return;
}
















